# Functions to calculate Zernike polynomials
#

import numpy as np
from numpy import sqrt
import functools
import sys

def NumZernike(m):
    """Return the number of polynomials up to and including radial order m"""
    return (m+1)*(m+2)//2

@functools.lru_cache()
def ZernikeGrid(gridSize,maxRadial,diameter=None):
    if diameter == None: diameter=gridSize
    radius=diameter/2.0
    x,y=np.mgrid[0:gridSize,0:gridSize]
    x = (x-(gridSize-1.0)/2.0)/radius
    y = (y-(gridSize-1.0)/2.0)/radius
    return ZernikeSequence(x,y,maxRadial)
    
def ZernikeSequence(x,y,maxRadial=5,epsilon=1e-10):
    """
    Return Zernike values at a given set of x,y coords up to some maximum 
    radial order
    """
    if maxRadial>5:
        raise ValueError('Code for higher radial orders not implemented')
    # Derive radius and exp(i*theta) 
    temp = x + 1j*y
    r1 = np.abs(temp)
    e1 = temp/(r1+epsilon)
     # Generate powers of r recursively
    r2 = r1*r1
    r3 = r2*r1
    r4 = r3*r1
    r5 = r4*r1
    # Generate cos and sin terms recursively from exp(i*theta)
    e2 = e1*e1
    e3 = e2*e1
    e4 = e3*e1
    e5 = e4*e1
    ctheta = e1.real
    stheta = e1.imag
    c2theta = e2.real
    s2theta = e2.imag
    c3theta = e3.real
    s3theta = e3.imag
    c4theta = e4.real
    s4theta = e4.imag
    c5theta = e5.real
    s5theta = e5.imag
    # Generate all the zernikes
    zernike = np.zeros((21,)+x.shape )
    zernike[0] =  1.0
    zernike[1] =  2.0*r1*ctheta
    zernike[2] =  2.0*r1*stheta
    zernike[3] =  sqrt(3.0)*(2.0*r2 - 1.0)
    zernike[4] =  sqrt(6.0)*r2*s2theta
    zernike[5] =  sqrt(6.0)*r2*c2theta
    zernike[6] =  sqrt(8.0)*(3.0*r3 - 2.0*r1)*stheta
    zernike[7] =  sqrt(8.0)*(3.0*r3 - 2.0*r1)*ctheta
    zernike[8] =  sqrt(8.0)*r3*s3theta
    zernike[9] = sqrt(8.0)*r3*c3theta
    zernike[10] = sqrt(5.0)*(6.*r4 - 6.*r2 + 1.)
    zernike[11] = sqrt(10.)*(4.*r4 - 3.*r2)*c2theta
    zernike[12] = sqrt(10.)*(4.*r4 - 3.*r2)*s2theta
    zernike[13] = sqrt(10.)*r4*c4theta
    zernike[14] = sqrt(10.)*r4*s4theta
    zernike[15] = sqrt(12.)*(10*r5-12*r3+3*r1)*ctheta
    zernike[16] = sqrt(12.)*(10*r5-12*r3+3*r1)*stheta
    zernike[17] = sqrt(12.)*(5*r5-4*r3)*c3theta
    zernike[18] = sqrt(12.)*(5*r5-4*r3)*s3theta
    zernike[19] = sqrt(12.)*r5*c5theta
    zernike[20] = sqrt(12.)*r5*s5theta
    # Make zernike zero outside unit circle (useful for dot product)
    zernike = zernike*np.less_equal(r1, 1.0)
    return(zernike[:NumZernike(maxRadial)])

if __name__ == '__main__':
    test()



import numpy as np
from numpy import fft, random
from numpy import add, sum, zeros, sqrt, arange 

def ScreenGenerator(nfft, r0, nx, ny):
    """Generate an infinite series of rectangular phase screens
    Uses an FFT screen generator to make a large screen and then
    returns non-overlapping subsections of it"""
    while 1:
        layers = GenerateTwoScreens(nfft, r0)
        for iLayer in range(2):
            for iy in range(int(nfft/ny)):
                for ix in range(int(nfft/nx)):
                    yield layers[iLayer][iy*ny:iy*ny+ny, ix*nx:ix*nx+nx]

def GenerateTwoScreens(nfft, r0):
    """

      Generate phase screens with a Kolmogorov spectrum of atmospheric
      disturbances [c.f. Tatarski 1961,1971], such that the phase structure
      function is given by

           D(r) = <[phi(r')-phi(r'+r)]**2>
                = 6.88*(r/r0)**5/3

       where r0 is the Fried parameter.

       This version returns two screens, because it's easier to do it that
       way.
    """
    C = sqrt(0.0229*(float(nfft)/r0)**(5.0/3.0))
    # Generate a 2-d array populated with rsquared=xsquared+ysquared
    r = arange(nfft)
    r[nfft/2:] = nfft-r[nfft/2:]
    rsq = r**2
    rsq = add.outer(rsq,rsq)
    rsq[0, 0] = 1.0 # To solve pole at origin problem
    sample = random.normal(size=(nfft, nfft))+1j*random.normal(size=(nfft, nfft))
    sample *= C*rsq**(-11.0/12.0)
    result = fft.fft2(sample)
    return(result.real, result.imag)

Noll=[1.0299,0.582,0.134,0.111,0.0880,0.0648,0.0587,0.0525,0.0463,0.0401,0.0377,0.0352,0.0328,0.0304,0.0279,0.0267,0.0255,0.0243,0.0232,0.0220,0.0208]


def NollVariance(screen, zernike):
    """ Return the values for the variance of a phase screen as increasing
    numbers of Zernikes are removed"""
    screen = screen*zernike[0] # Cut out a circle
    normalisation = sum(sum(zernike[0],-1))
    result = zeros(len(zernike))
    for i in range(len(zernike)):
        screen = screen - zernike[i]*sum(sum(zernike[i]*screen,-1))/normalisation
        result[i] = sum(sum(screen**2,-1))/normalisation
    return result

def test(diameter=32, screenSize=256, numIter=100):
    '''Test Noll theory on simulated turbulence'''
    import sys
    from Zernike15 import Zernike15

    zernike = Zernike15(diameter, diameter/2.0)
    screenGenerator = ScreenGenerator(screenSize, float(diameter),
                                      diameter, diameter)
    numScreen = 0
    variance = zeros(len(zernike))
    for i in range(numIter):
        screen = next(screenGenerator)
        variance = variance + NollVariance(screen, zernike)
    variance = variance/numIter
    print('Zernikes removed     Residual variance    Noll Result')
    for i in range(len(variance)):
        print('%16d %16f %16f' % (i, variance[i], Noll[i]))


if __name__ == '__main__':
    test(numIter=1000)



from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from builtins import *

import numpy as np
from .Zernike import ZernikeGrid
from .PhaseScreen import ScreenGenerator
import functools
import sys


__version__="0.2.0"

def npreshape(a, newShape):
    if newShape[0] == -1:
        if len(newShape) != 2:
            raise ValueError("new shape must be 2d")
        newShape=(np.prod(a.shape)/newShape[1],newShape[1])
    b=a.view()
    b.shape=newShape
    return b

def PhaseScreens(numTelescope,r0,gridSize,screenSize=1024):
    """
    Return a generator for atmospheric wavefront across a set of 
    numTelescope telescopes. Use next() to get the next set of screens.
    """
    screenGenerators=[ScreenGenerator(screenSize,r0,gridSize,gridSize)
                      for i in range(numTelescope)]
    while 1:
        yield np.array([next(screen) for screen in screenGenerators])

@functools.lru_cache()
def RadiusGrid(gridSize):
    """
    Return a square grid with values of the distance from the centre 
    of the grid to each gridpoint
    """
    x,y=np.mgrid[0:gridSize,0:gridSize]
    x = x-(gridSize-1.0)/2.0
    y = y-(gridSize-1.0)/2.0
    return np.abs(x+1j*y)

@functools.lru_cache()
def CircularMaskGrid(gridSize, diameter=None):
    """
    Return a square grid with ones inside and zeros outside a given 
    diameter circle
    """
    if diameter is None: diameter=gridSize
    return np.less_equal(RadiusGrid(gridSize),diameter/2.0)


def ComplexPupil(pupils,diameter=None):
    return np.exp(1j*pupils)*CircularMaskGrid(pupils.shape[-1],diameter)

def AdaptiveOpticsCorrect(pupils,diameter,maxRadial,numRemove=None):
    """
    Correct a wavefront using Zernike rejection up to some maximal order. 
    Can operate on multiple telescopes in parallel.
    Note that this version removes the piston mode as well
    """
    gridSize=pupils.shape[-1]
    pupilsVector=np.reshape(pupils,(-1,gridSize**2))
    zernikes=np.reshape(ZernikeGrid(gridSize,maxRadial,diameter),(-1,gridSize**2))
    if numRemove is None: numRemove=zernikes.shape[0]
    numScreen=pupilsVector.shape[0]
    normalisation=1.0/np.sum(zernikes[0])
    # Note extra iteration to remove residual piston
    for i in list(range(numRemove))+[0,]:
        amplitudes=np.inner(zernikes[i],pupilsVector)*normalisation
        pupilsVector=pupilsVector-zernikes[i]*amplitudes[:,np.newaxis]
    return np.reshape(pupilsVector,pupils.shape)


@functools.lru_cache()
def FibreMode(gridSize,modeDiameter):
    """
    Return a pupil-plane Gaussian mode with 1/e diameter given by 
    *modeDiameter*, normalised so that integral power over the mode is unity
    """
    rmode=modeDiameter/2
    return np.exp(-(RadiusGrid(gridSize)/rmode)**2)/(np.sqrt(np.pi/2)*rmode)

def FibreCouple(pupils,modeDiameter):
    """
    Return the complex amplitudes coupled into a set of fibers
    """
    gridSize=pupils.shape[-1]
    pupilsVector=np.reshape(pupils,(-1,gridSize**2))
    mode=np.reshape(FibreMode(gridSize,modeDiameter),(gridSize**2,))
    return np.inner(pupilsVector,mode)

def MonomodeCombine(pupils,modeDiameter=None):
    """
    Return the instantaneous coherent fluxes and photometric fluxes for a
    multiway monomode fibre combiner
    """
    if modeDiameter is None:
        modeDiameter=0.9*pupils.shape[-1]
    amplitudes=FibreCouple(pupils,modeDiameter)
    cc=np.conj(amplitudes)
    fluxes=(amplitudes*cc).real
    coherentFluxes=[amplitudes[i]*cc[j]
                    for i in range(1,len(amplitudes))
                    for j in range(i)]
    return fluxes,coherentFluxes

def MultimodeCombine(pupils):
    """
    Return the instantaneous coherent fluxes and photometric fluxes for a
    multiway multimode combiner (no spatial filtering)
    """
    fluxes=[np.vdot(pupils[i],pupils[i]).real for i in range(len(pupils))]
    coherentFluxes=[np.vdot(pupils[i],pupils[j])
                    for i in range(1,len(pupils))
                    for j in range(i)]
    return fluxes,coherentFluxes




from __future__ import print_function
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand
import io
import codecs
import os
import sys

import pois

here = os.path.abspath(os.path.dirname(__file__))

def read(*filenames, **kwargs):
    encoding = kwargs.get('encoding', 'utf-8')
    sep = kwargs.get('sep', '\n')
    buf = []
    for filename in filenames:
        with io.open(filename, encoding=encoding) as f:
            buf.append(f.read())
    return sep.join(buf)

long_description = read('README.rst', 'CHANGES.txt')

# Use pytest for testing
class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        #import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)

setup(
    name='pois',
    version=pois.__version__,
    url='http://github.com/dbuscher/pois/',
    author='David Buscher',
    install_requires=['numpy',
                    ],
    author_email='dfb@mrao.cam.ac.uk',
    description='Python optical interferometer simulation',
    long_description=long_description,
    packages=['pois'],
    include_package_data=True,
    platforms='any',
    classifiers = [
        'Programming Language :: Python :: 3',
        'Development Status :: 4 - Beta',
        'Natural Language :: English',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Topic :: Scientific/Engineering :: Astronomy',
        'Topic :: Scientific/Engineering :: Physics',
        ],
    tests_require=['pytest'],
    cmdclass = {'test': PyTest},
)
